@@ -2,7 +2,7 @@
  *
  *  BlueZ - Bluetooth protocol stack for Linux
  *
- *  Copyright (C) 2004-2008  Marcel Holtmann <marcel@holtmann.org>
+ *  Copyright (C) 2004-2009  Marcel Holtmann <marcel@holtmann.org>
  *
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -196,7 +196,7 @@ static int network_probe(struct btd_device *device, GSList *uuids, uint16_t id)
 	adapter_get_address(adapter, &src);
 	device_get_address(device, &dst);
 
-	return connection_register(path, &src, &dst, id);
+	return connection_register(device, path, &src, &dst, id);
 }
 
 static void network_remove(struct btd_device *device, uint16_t id)
